import hashlib
import random
import re
import torch

from itertools import chain
from collections.abc import Mapping

# See https://www.drive5.com/usearch/manual/IUPAC_codes.html
nucleotide_wildcard_mapping = {
    'R': ['A', 'G'],
    'Y': ['C', 'T'],
    'S': ['G', 'C'],
    'W': ['A', 'T'],
    'K': ['G', 'T'],
    'M': ['A', 'C'],
    'B': ['C', 'G', 'T'],
    'D': ['A', 'G', 'T'],
    'H': ['A', 'C', 'T'],
    'V': ['A', 'C', 'G']
}
nucleotide_wildcards = set(nucleotide_wildcard_mapping)
nucleotide_wildcard_re = re.compile('(' + '|'.join(nucleotide_wildcards) + ')')


def identity(x, *args, **kwargs):
    return x


def merge_dicts(dicts, cat_fn=None):
    """
    Args:
        dicts: Dictionaries to merge key-wise.
        cat_fn (optional): Function to use for merging dictionary values.

    Example:
    >>> merge_dicts({'a': 0, 'b': 42}, {'a': 1})
    {'a': [0, 1], 'b': [42]}
    """
    if cat_fn is None:
        cat_fn = identity

    keys = set(chain.from_iterable(dicts))
    merged = {k: cat_fn([d[k] for d in dicts if k in d]) for k in keys}
    return merged


def digest_strings(ss):
    m = hashlib.md5()
    for s in ss:
        m.update(s.encode())
    return m.hexdigest()


def get_lmdb_keys(env):
    with env.begin() as txn:
        return list(txn.cursor().iternext(values=False))


def to_device(o, device=None):
    if isinstance(o, torch.Tensor):
        return o.to(device)
    if isinstance(o, (list, tuple)):
        return type(o)(to_device(el, device) for el in o)
    if isinstance(o, Mapping):
        return {k: to_device(v, device) for k, v in o.items()}
    return o


def sub_nucleotide_wildcards(seq):
    res = seq
    for match in nucleotide_wildcard_re.finditer(seq):
        res = res[:match.start()] + random.choice(nucleotide_wildcard_mapping[match.group()]) + res[match.end():]
    return res



import torch
import torch.nn as nn

from collections import OrderedDict
from typing import Dict, Callable
from enum import Enum
from transformers import (
    BertModel,
    BertTokenizer,
    T5EncoderModel,
    T5Tokenizer
)


class EmbeddingKind(Enum):
    PER_RESIDUE = 'per_residue'
    PER_PROTEIN = 'per_protein'


class BaseProteinEmbeddingModel(nn.Module):
    embedding_kind: EmbeddingKind

    def prepare_sequences(self, sequences):
        return NotImplementedError

    def forward(self, input):
        raise NotImplementedError


def load_huggingface_language_model(model_cls, tokenizer_cls, model_name, load_weights=True):
    if load_weights:
        model = model_cls.from_pretrained(model_name),
        tokenizer = tokenizer_cls.from_pretrained(model_name)
        return model, tokenizer
    else:
        config = model_cls.config_class.from_pretrained(model_name)
        model = model_cls(config)
        tokenizer = tokenizer_cls.from_pretrained(model_name)
        return model, tokenizer


class BaseProtTransEmbeddingModel(BaseProteinEmbeddingModel):
    embedding_kind = EmbeddingKind.PER_RESIDUE
    available_models = None

    def __init__(self, model, tokenizer):
        super().__init__()
        self.model = model
        self.tokenizer = tokenizer

    def _validate_model_name(self, model_name):
        assert self.available_models is None or model_name in self.available_models, \
            f'Unknown model name \'{model_name}\'. Available options are {self.available_models}'

    def prepare_sequences(self, sequences):
        return self.tokenizer.batch_encode_plus(
            sequences,
            return_tensors='pt',
            add_special_tokens=True,
            padding=True
        )

    def _post_process_embedding(self, embed, seq_len):
        raise NotImplementedError

    @torch.no_grad()
    def forward(self,  input):
        attn_mask = input['attention_mask']

        output = self.model(**input)

        embeddings = output.last_hidden_state.cpu()
        seq_lens = (attn_mask == 1).sum(-1)

        for embed, seq_len in zip(embeddings, seq_lens):
            # Tokenized sequences have the following form:
            # [CLS] V N ... I K [SEP] [PAD] ... [PAD]
            #
            # We remove the special tokens ([CLS], [SEP], [PAD]) before
            # computing the mean over the remaining sequence

            yield self._post_process_embedding(embed, seq_len)


class ProtBERTEmbeddingModel(BaseProtTransEmbeddingModel):
    available_models = [
        'prot_bert',
        'prot_bert_bfd'
    ]

    def __init__(self, model_name: str, load_weights: bool = True):
        self._validate_model_name(model_name)

        mode_name = f'Rostlab/{model_name}'
        model, tokenizer = load_huggingface_language_model(
            BertModel,
            BertTokenizer,
            mode_name,
            load_weights=load_weights
        )

        super().__init__(model=model, tokenizer=tokenizer)

    def _post_process_embedding(self, embed, seq_len):
        return embed[1:seq_len - 1]


class ProtT5EmbeddingModel(BaseProtTransEmbeddingModel):
    available_models = [
        'prot_t5_xl_uniref50',
        'prot_t5_xl_bfd',
        'prot_t5_xxl_uniref50',
        'prot_t5_xxl_bfd'
    ]

    def __init__(self, model_name: str, load_weights: bool = True):
        self._validate_model_name(model_name)

        mode_name = f'Rostlab/{model_name}'
        model, tokenizer = load_huggingface_language_model(
            T5EncoderModel,
            T5Tokenizer,
            mode_name,
            load_weights=load_weights
        )

        super().__init__(model=model, tokenizer=tokenizer)

    def _post_process_embedding(self, embedding, seq_len):
        return embedding[:seq_len - 1]


class ESMEmbeddingModel(BaseProteinEmbeddingModel):
    embedding_kind = EmbeddingKind.PER_RESIDUE

    def __init__(self, model_name: str, repr_layer: int):
        super().__init__()

        self.model, self.alphabet = torch.hub.load('facebookresearch/esm:main', model_name)
        self.batch_converter = self.alphabet.get_batch_converter()
        self.repr_layer = repr_layer

    def prepare_sequences(self, sequences):
        _, _, batch_tokens = self.batch_converter([('', seq) for seq in sequences])

        return {
            'tokens': batch_tokens
        }

    @torch.no_grad()
    def forward(self, input):
        results = self.model(**input, repr_layers=[self.repr_layer], return_contacts=True)
        token_representations = results['representations'][self.repr_layer]
        seq_lengths = (input['tokens'] != self.alphabet.padding_idx).sum(1)

        for i, seq_len in enumerate(seq_lengths):
            yield token_representations[i, 1:seq_len - 1]


embedding_models: Dict[str, Callable[[], BaseProteinEmbeddingModel]] = OrderedDict([
    # ProtTrans family. See https://github.com/agemagician/ProtTrans
    ('prot_t5_xl_uniref50', lambda: ProtT5EmbeddingModel('prot_t5_xl_uniref50')),
    ('prot_t5_xl_bfd', lambda: ProtT5EmbeddingModel('prot_t5_xl_bfd')),
    ('prot_t5_xxl_uniref50', lambda: ProtT5EmbeddingModel('prot_t5_xxl_uniref50')),
    ('prot_t5_xxl_bfd', lambda: ProtT5EmbeddingModel('prot_t5_xxl_bfd')),
    ('prot_bert_bfd', lambda: ProtBERTEmbeddingModel('prot_bert_bfd')),
    ('prot_bert', lambda: ProtBERTEmbeddingModel('prot_bert')),

    # ESM family. See https://github.com/facebookresearch/esm
    ('esm2_t48_15B_UR50D', lambda: ESMEmbeddingModel('esm2_t48_15B_UR50D', repr_layer=48)),
    ('esm2_t36_3B_UR50D', lambda: ESMEmbeddingModel('esm2_t36_3B_UR50D', repr_layer=36)),
    ('esm2_t33_650M_UR50D', lambda: ESMEmbeddingModel('esm2_t33_650M_UR50D', repr_layer=33)),
    ('esm2_t30_150M_UR50D', lambda: ESMEmbeddingModel('esm2_t30_150M_UR50D', repr_layer=30)),
    ('esm2_t12_35M_UR50D', lambda: ESMEmbeddingModel('esm2_t12_35M_UR50D', repr_layer=12)),
    ('esm2_t6_8M_UR50D', lambda: ESMEmbeddingModel('esm2_t6_8M_UR50D', repr_layer=6))
])


def list_models():
    return list(embedding_models)


def get_model(model_name):
    return embedding_models[model_name]()



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['protenc']

package_data = \
{'': ['*']}

install_requires = \
['biopython>=1.80,<2.0',
 'lmdb>=1.3.0,<2.0.0',
 'pandas>=1.5.2,<2.0.0',
 'sentencepiece>=0.1.97,<0.2.0',
 'torch>=1.13.0,<2.0.0',
 'tqdm>=4.64.1,<5.0.0',
 'transformers>=4.24.0,<5.0.0']

setup_kwargs = {
    'name': 'protenc',
    'version': '0.1.0',
    'description': 'Simplify extraction of protein embedding from various models.',
    'long_description': "protenc\n=======\n\nprotenc is a library to simplify extraction of protein embeddings from various pre-trained models, including:\n\n* [ProtTrans](https://github.com/agemagician/ProtTrans) family\n* [ESM](https://github.com/facebookresearch/esm)\n* AlphaFold (coming soon™)\n\nIt provides a programmatic Python API as well as a highly flexible bulk extraction script, supporting many input and\noutput formats.\n\n**Note:** the project is work in progress.\n\nUsage\n-----\n\n**Installation**\n\n```bash\npip install protenc\n```\n\n**Python API**\n\n```python\nimport protenc\nimport torch\n\n# List available models\nprint(protenc.list_models())\n\n# Instantiate a model\nmodel = protenc.get_model('esm2_t33_650M_UR50D')\n\nproteins = [\n  'MKTVRQERLKSIVRILERSKEPVSGAQLAEELSVSRQVIVQDIAYLRSLGYNIVATPRGYVLAGG',\n  'KALTARQQEVFDLIRDHISQTGMPPTRAEIAQRLGFRSPNAAEEHLKALARKGVIEIVSGASRGIRLLQEE'\n]\n\nbatch = model.prepare_sequences(proteins)\n\n# Move to GPU if available\nif torch.cuda.is_available():\n  model = model.to('cuda')\n  batch = protenc.utils.to_device(batch, 'cuda')\n\nfor embed in model(batch):\n  # Embeddings have shape [L, D] where L is the sequence length and D the \n  # embedding dimensionality.\n  print(embed.shape)\n  \n  # Derive a single per-protein embedding vector by averaging along the \n  # sequence dimension\n  embed.mean(0)\n```\n\n**Command-line interface**\n\nComing soon.\n\nDevelopment\n-----------\n\nClone the repository:\n\n```bash\ngit clone git+https://github.com/kklemon/protenc.git\n```\n\nInstall dependencies via [Poetry](https://python-poetry.org/):\n\n```bash\npoetry install\n```\n\nTodo\n----\n\n- [ ] Support for more input formats\n  - [X] CSV\n  - [ ] Parquet\n  - [ ] FASTA\n  - [ ] JSON\n- [ ] Support for more output formats\n  - [X] LMDB\n  - [ ] HDF5\n  - [ ] DataFrame\n  - [ ] Pickle\n- [ ] Large models support\n  - [ ] Model offloading\n  - [ ] Sharding\n- [ ] Support for more protein language models\n  - [ ] While ProtTrans family\n  - [ ] While ESM family\n    - [ ] AlphaFold (?)\n- [ ] Implement all remaining TODOs in code\n- [ ] Distributed inference\n- [ ] Maybe support some sort of optimized inference such as quantization\n  - This may be up to the model providers\n- [ ] Improve documentation\n- [ ] Support translation of gene sequences\n",
    'author': 'Kristian Klemon',
    'author_email': 'kristian.klemon@gmail.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'None',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)
